@@ -125,7 +125,7 @@ class TensorDtype(pd.api.extensions.ExtensionDtype):
 
         >>> # In addition to doing Pandas operations on the tensor column,
         >>> # you can now put the DataFrame into a Dataset.
-        >>> ds = ray.data.from_pandas([ray.put(df)])
+        >>> ds = ray.data.from_pandas(df)
         >>> # Internally, this column is represented the corresponding
         >>> # Arrow tensor extension type.
         >>> ds.schema()
@@ -140,7 +140,7 @@ class TensorDtype(pd.api.extensions.ExtensionDtype):
         one: int64
         two: extension<arrow.py_extension_type<ArrowTensorType>>
 
-        >>> read_df = ray.get(read_ds.to_pandas())[0]
+        >>> read_df = ray.get(read_ds.to_pandas_refs())[0]
         >>> read_df.dtypes
         one          int64
         two    TensorDtype
@@ -407,7 +407,7 @@ class TensorArray(pd.api.extensions.ExtensionArray, TensorOpsMixin):
 
         >>> # In addition to doing Pandas operations on the tensor column,
         >>> # you can now put the DataFrame into a Dataset.
-        >>> ds = ray.data.from_pandas([ray.put(df)])
+        >>> ds = ray.data.from_pandas(df)
         >>> # Internally, this column is represented the corresponding
         >>> # Arrow tensor extension type.
         >>> ds.schema()
@@ -422,7 +422,7 @@ class TensorArray(pd.api.extensions.ExtensionArray, TensorOpsMixin):
         one: int64
         two: extension<arrow.py_extension_type<ArrowTensorType>>
 
-        >>> read_df = ray.get(read_ds.to_pandas())[0]
+        >>> read_df = ray.get(read_ds.to_pandas_refs())[0]
         >>> read_df.dtypes
         one          int64
         two    TensorDtype
@@ -1155,6 +1155,10 @@ class ArrowTensorType(pa.PyExtensionType):
         """
         return ArrowTensorArray
 
+    def __str__(self):
+        return "<ArrowTensorType: shape={}, dtype={}>".format(
+            self.shape, self.storage_type.value_type)
+
 
 @PublicAPI(stability="beta")
 class ArrowTensorArray(pa.ExtensionArray):
@@ -1230,6 +1234,12 @@ class ArrowTensorArray(pa.ExtensionArray):
                                      if element_shape else 1)
 
             # Data buffer.
+            if pa.types.is_boolean(pa_dtype):
+                # NumPy doesn't represent boolean arrays as bit-packed, so we manually
+                # bit-pack the booleans before handing the buffer off to Arrow.
+                # NOTE: Arrow expects LSB bit-packed ordering.
+                # NOTE: This creates a copy.
+                arr = np.packbits(arr, bitorder="little")
             data_buffer = pa.py_buffer(arr)
             data_array = pa.Array.from_buffers(pa_dtype, total_num_items,
                                                [None, data_buffer])
@@ -1250,7 +1260,9 @@ class ArrowTensorArray(pa.ExtensionArray):
         else:
             raise ValueError("Must give ndarray or iterable of ndarrays.")
 
-    def _to_numpy(self, index: Optional[int] = None):
+    def _to_numpy(self,
+                  index: Optional[int] = None,
+                  zero_copy_only: bool = False):
         """
         Helper for getting either an element of the array of tensors as an
         ndarray, or the entire array of tensors as a single ndarray.
@@ -1259,26 +1271,40 @@ class ArrowTensorArray(pa.ExtensionArray):
             index: The index of the tensor element that we wish to return as
                 an ndarray. If not given, the entire array of tensors is
                 returned as an ndarray.
+            zero_copy_only: If True, an exception will be raised if the
+                conversion to a NumPy array would require copying the
+                underlying data (e.g. in presence of nulls, or for
+                non-primitive types). This argument is currently ignored, so
+                zero-copy isn't enforced even if this argument is true.
 
         Returns:
             The corresponding tensor element as an ndarray if an index was
             given, or the entire array of tensors as an ndarray otherwise.
         """
+        # TODO(Clark): Enforce zero_copy_only.
+        # TODO(Clark): Support strides?
         # Buffers schema:
         # [None, offset_buffer, None, data_buffer]
         buffers = self.buffers()
         data_buffer = buffers[3]
         storage_list_type = self.storage.type
-        ext_dtype = storage_list_type.value_type.to_pandas_dtype()
+        value_type = storage_list_type.value_type
+        ext_dtype = value_type.to_pandas_dtype()
         shape = self.type.shape
-        # Size in bytes of the underlying ndarray item.
-        item_byte_width = storage_list_type.value_type.bit_width // 8
+        if pa.types.is_boolean(value_type):
+            # Arrow boolean array buffers are bit-packed, with 8 entries per byte,
+            # and are accessed via bit offsets.
+            buffer_item_width = value_type.bit_width
+        else:
+            # We assume all other array types are accessed via byte array
+            # offsets.
+            buffer_item_width = value_type.bit_width // 8
         # Number of items per inner ndarray.
         num_items_per_element = np.prod(shape) if shape else 1
         # Base offset into data buffer, e.g. due to zero-copy slice.
         buffer_offset = self.offset * num_items_per_element
-        # Offset (in bytes) of array data in buffer.
-        offset = item_byte_width * buffer_offset
+        # Offset of array data in buffer.
+        offset = buffer_item_width * buffer_offset
         if index is not None:
             # Getting a single tensor element of the array.
             offset_buffer = buffers[1]
@@ -1287,19 +1313,50 @@ class ArrowTensorArray(pa.ExtensionArray):
             # Offset into array to reach logical index.
             index_offset = offset_array[index]
             # Add the index offset to the base offset.
-            offset += item_byte_width * index_offset
+            offset += buffer_item_width * index_offset
         else:
             # Getting the entire array of tensors.
             shape = (len(self), ) + shape
-        # TODO(Clark): Support strides?
+        if pa.types.is_boolean(value_type):
+            # Special handling for boolean arrays, since Arrow bit-packs boolean arrays
+            # while NumPy does not.
+            # Cast as uint8 array and let NumPy unpack into a boolean view.
+            # Offset into uint8 array, where each element is a bucket for 8 booleans.
+            byte_bucket_offset = offset // 8
+            # Offset for a specific boolean, within a uint8 array element.
+            bool_offset = offset % 8
+            # The number of uint8 array elements (buckets) that our slice spans.
+            # Note that, due to the offset for a specific boolean, the slice can span
+            # byte boundaries even if it contains less than 8 booleans.
+            num_boolean_byte_buckets = 1 + (
+                (bool_offset + np.prod(shape) - 1) // 8)
+            # Construct the uint8 array view on the buffer.
+            arr = np.ndarray(
+                (num_boolean_byte_buckets, ),
+                dtype=np.uint8,
+                buffer=data_buffer,
+                offset=byte_bucket_offset,
+            )
+            # Unpack into a byte per boolean, using LSB bit-packed ordering.
+            arr = np.unpackbits(arr, bitorder="little")
+            # Interpret buffer as boolean array.
+            return np.ndarray(
+                shape, dtype=np.bool_, buffer=arr, offset=bool_offset)
         return np.ndarray(
             shape, dtype=ext_dtype, buffer=data_buffer, offset=offset)
 
-    def to_numpy(self):
+    def to_numpy(self, zero_copy_only: bool = True):
         """
         Convert the entire array of tensors into a single ndarray.
 
+        Args:
+            zero_copy_only: If True, an exception will be raised if the
+                conversion to a NumPy array would require copying the
+                underlying data (e.g. in presence of nulls, or for
+                non-primitive types). This argument is currently ignored, so
+                zero-copy isn't enforced even if this argument is true.
+
         Returns:
             A single ndarray representing the entire array of tensors.
         """
-        return self._to_numpy()
+        return self._to_numpy(zero_copy_only=zero_copy_only)
