@@ -28,7 +28,7 @@ from ray.serve.handle import RayServeHandle, RayServeSyncHandle
 from ray.serve.http_util import ASGIHTTPSender, make_fastapi_class_based_view
 from ray.serve.utils import (LoggingContext, ensure_serialization_context,
                              format_actor_name, get_current_node_resource_key,
-                             get_random_letters, logger)
+                             get_random_letters, logger, DEFAULT)
 from ray.util.annotations import PublicAPI
 import ray
 from ray import cloudpickle
@@ -71,6 +71,7 @@ class ReplicaContext:
     deployment: str
     replica_tag: ReplicaTag
     _internal_controller_name: str
+    _internal_controller_namespace: str
     servable_object: Callable
 
 
@@ -78,11 +79,13 @@ def _set_internal_replica_context(
         deployment: str,
         replica_tag: ReplicaTag,
         controller_name: str,
+        controller_namespace: str,
         servable_object: Callable,
 ):
     global _INTERNAL_REPLICA_CONTEXT
     _INTERNAL_REPLICA_CONTEXT = ReplicaContext(
-        deployment, replica_tag, controller_name, servable_object)
+        deployment, replica_tag, controller_name, controller_namespace,
+        servable_object)
 
 
 def _ensure_connected(f: Callable) -> Callable:
@@ -106,7 +109,9 @@ class Client:
         self._shutdown = False
         self._http_config: HTTPOptions = ray.get(
             controller.get_http_config.remote())
-        self._root_url = ray.get(self._controller.get_root_url.remote())
+        self._root_url = ray.get(controller.get_root_url.remote())
+        self._checkpoint_path = ray.get(
+            controller.get_checkpoint_path.remote())
 
         # Each handle has the overhead of long poll client, therefore cached.
         self.handle_cache = dict()
@@ -126,6 +131,14 @@ class Client:
     def root_url(self):
         return self._root_url
 
+    @property
+    def http_config(self):
+        return self._http_config
+
+    @property
+    def checkpoint_path(self):
+        return self._checkpoint_path
+
     def __del__(self):
         if not self._detached:
             logger.debug("Shutting down Ray Serve because client went out of "
@@ -198,7 +211,7 @@ class Client:
             version: Optional[str] = None,
             prev_version: Optional[str] = None,
             route_prefix: Optional[str] = None,
-            url: str = "",
+            url: Optional[str] = None,
             _blocking: Optional[bool] = True) -> Optional[GoalId]:
         if config is None:
             config = {}
@@ -226,6 +239,13 @@ class Client:
             raise TypeError(
                 "config must be a DeploymentConfig or a dictionary.")
 
+        if deployment_config.autoscaling_config is not None and \
+            deployment_config.max_concurrent_queries < deployment_config. \
+                autoscaling_config.target_num_ongoing_requests_per_replica:
+            logger.warning("Autoscaling will never happen, "
+                           "because 'max_concurrent_queries' is less than "
+                           "'target_num_ongoing_requests_per_replica' now.")
+
         goal_id, updating = ray.get(
             self._controller.deploy.remote(name,
                                            deployment_config.to_proto_bytes(),
@@ -246,9 +266,14 @@ class Client:
 
         if _blocking:
             self._wait_for_goal(goal_id)
+
+            if url is not None:
+                url_part = f" at `{url}`"
+            else:
+                url_part = ""
             logger.info(
                 f"Deployment '{name}{':'+version if version else ''}' is ready"
-                f" at `{url}`. {tag}")
+                f"{url_part}. {tag}")
         else:
             return goal_id
 
@@ -268,17 +293,17 @@ class Client:
     @_ensure_connected
     def get_handle(
             self,
-            endpoint_name: str,
+            deployment_name: str,
             missing_ok: Optional[bool] = False,
             sync: bool = True,
             _internal_pickled_http_request: bool = False,
     ) -> Union[RayServeHandle, RayServeSyncHandle]:
-        """Retrieve RayServeHandle for service endpoint to invoke it from Python.
+        """Retrieve RayServeHandle for service deployment to invoke it from Python.
 
         Args:
-            endpoint_name (str): A registered service endpoint.
-            missing_ok (bool): If true, then Serve won't check the endpoint is
-                registered. False by default.
+            deployment_name (str): A registered service deployment.
+            missing_ok (bool): If true, then Serve won't check the deployment
+                is registered. False by default.
             sync (bool): If true, then Serve will return a ServeHandle that
                 works everywhere. Otherwise, Serve will return a ServeHandle
                 that's only usable in asyncio loop.
@@ -286,15 +311,15 @@ class Client:
         Returns:
             RayServeHandle
         """
-        cache_key = (endpoint_name, missing_ok, sync)
+        cache_key = (deployment_name, missing_ok, sync)
         if cache_key in self.handle_cache:
             cached_handle = self.handle_cache[cache_key]
             if cached_handle.is_polling and cached_handle.is_same_loop:
                 return cached_handle
 
         all_endpoints = ray.get(self._controller.get_all_endpoints.remote())
-        if not missing_ok and endpoint_name not in all_endpoints:
-            raise KeyError(f"Endpoint '{endpoint_name}' does not exist.")
+        if not missing_ok and deployment_name not in all_endpoints:
+            raise KeyError(f"Deployment '{deployment_name}' does not exist.")
 
         try:
             asyncio_loop_running = asyncio.get_event_loop().is_running()
@@ -322,13 +347,13 @@ class Client:
         if sync:
             handle = RayServeSyncHandle(
                 self._controller,
-                endpoint_name,
+                deployment_name,
                 _internal_pickled_http_request=_internal_pickled_http_request,
             )
         else:
             handle = RayServeHandle(
                 self._controller,
-                endpoint_name,
+                deployment_name,
                 _internal_pickled_http_request=_internal_pickled_http_request,
             )
 
@@ -353,6 +378,35 @@ class Client:
         return handle
 
 
+def _check_http_and_checkpoint_options(
+        client: Client,
+        http_options: Union[dict, HTTPOptions],
+        checkpoint_path: str,
+) -> None:
+    if checkpoint_path and checkpoint_path != client.checkpoint_path:
+        logger.warning(
+            f"The new client checkpoint path '{checkpoint_path}' "
+            f"is different from the existing one '{client.checkpoint_path}'. "
+            "The new checkpoint path is ignored.")
+
+    if http_options:
+        client_http_options = client.http_config
+        new_http_options = http_options if isinstance(
+            http_options, HTTPOptions) else HTTPOptions.parse_obj(http_options)
+        different_fields = []
+        all_http_option_fields = new_http_options.__dict__
+        for field in all_http_option_fields:
+            if getattr(new_http_options, field) != getattr(
+                    client_http_options, field):
+                different_fields.append(field)
+
+        if len(different_fields):
+            logger.warning(
+                "The new client HTTP config differs from the existing one "
+                f"in the following fields: {different_fields}. "
+                "The new HTTP config is ignored.")
+
+
 @PublicAPI(stability="beta")
 def start(
         detached: bool = False,
@@ -413,6 +467,10 @@ def start(
         client = _get_global_client()
         logger.info("Connecting to existing Serve instance in namespace "
                     f"'{controller_namespace}'.")
+
+        _check_http_and_checkpoint_options(client, http_options,
+                                           _checkpoint_path)
+
         return client
     except RayServeException:
         pass
@@ -487,7 +545,7 @@ def _connect() -> Client:
         controller_namespace = _get_controller_namespace(detached=True)
     else:
         controller_name = _INTERNAL_REPLICA_CONTEXT._internal_controller_name
-        controller_namespace = _get_controller_namespace(detached=False)
+        controller_namespace = _INTERNAL_REPLICA_CONTEXT._internal_controller_namespace
 
     # Try to get serve controller if it exists
     try:
@@ -635,7 +693,7 @@ class Deployment:
                  prev_version: Optional[str] = None,
                  init_args: Optional[Tuple[Any]] = None,
                  init_kwargs: Optional[Tuple[Any]] = None,
-                 route_prefix: Optional[str] = None,
+                 route_prefix: Union[str, None, DEFAULT] = DEFAULT.VALUE,
                  ray_actor_options: Optional[Dict] = None,
                  _internal=False) -> None:
         """Construct a Deployment. CONSTRUCTOR SHOULDN'T BE USED DIRECTLY.
@@ -662,7 +720,7 @@ class Deployment:
             raise TypeError("init_args must be a tuple.")
         if not (init_kwargs is None or isinstance(init_kwargs, dict)):
             raise TypeError("init_kwargs must be a dict.")
-        if route_prefix is not None:
+        if route_prefix is not DEFAULT.VALUE and route_prefix is not None:
             if not isinstance(route_prefix, str):
                 raise TypeError("route_prefix must be a string.")
             if not route_prefix.startswith("/"):
@@ -744,6 +802,8 @@ class Deployment:
     @property
     def route_prefix(self) -> Optional[str]:
         """HTTP route prefix that this deployment is exposed under."""
+        if self._route_prefix is DEFAULT.VALUE:
+            return f"/{self._name}"
         return self._route_prefix
 
     @property
@@ -762,10 +822,13 @@ class Deployment:
         return self._init_args
 
     @property
-    def url(self):
+    def url(self) -> Optional[str]:
         """Full HTTP url for this deployment."""
-        return _get_global_client().root_url + (self._route_prefix
-                                                or f"/{self._name}")
+        if self._route_prefix is None:
+            # this deployment is not exposed over HTTP
+            return None
+
+        return _get_global_client().root_url + self.route_prefix
 
     def __call__(self):
         raise RuntimeError("Deployments cannot be constructed directly. "
@@ -795,7 +858,7 @@ class Deployment:
             config=self._config,
             version=self._version,
             prev_version=self._prev_version,
-            route_prefix=self._route_prefix,
+            route_prefix=self.route_prefix,
             url=self.url,
             _blocking=_blocking)
 
@@ -829,7 +892,7 @@ class Deployment:
                 prev_version: Optional[str] = None,
                 init_args: Optional[Tuple[Any]] = None,
                 init_kwargs: Optional[Dict[Any, Any]] = None,
-                route_prefix: Optional[str] = None,
+                route_prefix: Union[str, None, DEFAULT] = DEFAULT.VALUE,
                 num_replicas: Optional[int] = None,
                 ray_actor_options: Optional[Dict] = None,
                 user_config: Optional[Any] = None,
@@ -867,16 +930,14 @@ class Deployment:
         if init_kwargs is None:
             init_kwargs = self._init_kwargs
 
-        if route_prefix is None:
-            if self._route_prefix == f"/{self._name}":
-                route_prefix = None
-            else:
-                route_prefix = self._route_prefix
+        if route_prefix is DEFAULT.VALUE:
+            # Default is to keep the previous value
+            route_prefix = self._route_prefix
 
         if ray_actor_options is None:
             ray_actor_options = self._ray_actor_options
 
-        if _autoscaling_config is None:
+        if _autoscaling_config is not None:
             new_config.autoscaling_config = _autoscaling_config
 
         if _graceful_shutdown_wait_loop_s is not None:
@@ -907,18 +968,15 @@ class Deployment:
             self._config == other._config,
             self._init_args == other._init_args,
             self._init_kwargs == other._init_kwargs,
-            self._route_prefix == other._route_prefix,
+            # compare route prefix with default value resolved
+            self.route_prefix == other.route_prefix,
             self._ray_actor_options == self._ray_actor_options,
         ])
 
     def __str__(self):
-        if self._route_prefix is None:
-            route_prefix = f"/{self._name}"
-        else:
-            route_prefix = self._route_prefix
         return (f"Deployment(name={self._name},"
                 f"version={self._version},"
-                f"route_prefix={route_prefix})")
+                f"route_prefix={self.route_prefix})")
 
     def __repr__(self):
         return str(self)
@@ -956,7 +1014,7 @@ def deployment(
         num_replicas: Optional[int] = None,
         init_args: Optional[Tuple[Any]] = None,
         init_kwargs: Optional[Dict[Any, Any]] = None,
-        route_prefix: Optional[str] = None,
+        route_prefix: Union[str, None, DEFAULT] = DEFAULT.VALUE,
         ray_actor_options: Optional[Dict] = None,
         user_config: Optional[Any] = None,
         max_concurrent_queries: Optional[int] = None,
@@ -989,6 +1047,7 @@ def deployment(
             passed when you call `.deploy()` on the returned Deployment.
         route_prefix (Optional[str]): Requests to paths under this HTTP path
             prefix will be routed to this deployment. Defaults to '/{name}'.
+            When set to 'None', no HTTP endpoint will be created.
             Routing is done based on longest-prefix match, so if you have
             deployment A with a prefix of '/a' and deployment B with a prefix
             of '/a/b', requests to '/a', '/a/', and '/a/c' go to A and requests
